"""
Flow exceptions module
"""


class FlowException(Exception):
    """
    Class custom exception
    for incorrect type flow
    """

    def __init__(self, message: str):
        self.message = f"The flow chain ended without finding a single handler: {message}"
        Exception.__init__(self, self.message)



"""
Module with base class blocks
"""
from __future__ import annotations

import types
from abc import ABC, abstractmethod
from typing import Optional, Callable

from orchestrator_service.message import Message
from .exc import FlowException


class BlockHandler(ABC):
    """
    The Handler interface declares a method for building a chain of handlers.
    It also declares a method to fulfill the request.
    """

    @abstractmethod
    def set_next(self,
                 handler: BlockHandler) -> BlockHandler:
        """
        method for adding a new handler
        :param BlockHandler handler: object next handler in chain flow
        :return: BlockHandler
        """
        raise NotImplementedError

    @abstractmethod
    def handle(self,
               message: Message) -> Optional[Exception]:
        """
        flow chain management method
        :param MessageQueue message:
        :return: None
        """
        raise NotImplementedError

    @abstractmethod
    def process(self,
                message: Message) -> None:
        """
        method for executing the logic of a given block
        in it, only send messages to other services
        :param MessageQueue message: msg to process
        :return: None
        """
        raise NotImplementedError

    @abstractmethod
    def get_list_flow(self) -> str:
        """
        Method return str steps flow
        :return: str
        """
        raise NotImplementedError


class Block(BlockHandler):
    """
    The main class for inheriting the blocks that make up the flow of tasks execution
    """
    _next_handler: BlockHandler = None
    # if the function returns None then you do not need to perform the following steps
    _pre_handler_function: Callable = None
    _post_handler_function: Callable = None

    @property
    def pre_handler_function(self):
        """
        function which call before send to handler
        :return:
        """
        return self._pre_handler_function

    @property
    def post_handler_function(self):
        """
        function which call after received from source
        :return:
        """
        return self._post_handler_function

    @property
    def name_block(self):
        """
        Unique name to identify block
        for override in subclass   name_block
        """
        raise NotImplementedError

    # @property
    # def name_queue(self):
    #     """
    #     Name of the queue that the service is listening on for this block
    #     for override in subclass name_queue
    #     """
    #     raise NotImplementedError

    def __init__(self,
                 pre_handler_function: Callable = None,
                 post_handler_function: Callable = None):
        """
        Init Block
        :param pre_handler_function: function should accept and return objects of type Message
        which be run before call method
        :param post_handler_function: function should accept and return objects of type Message
        which be run after got msg from source

        """
        self.pre_handler_function = pre_handler_function
        self.post_handler_function = post_handler_function

    def set_next(self,
                 handler: BlockHandler) -> Optional[BlockHandler, None]:
        """
        Save Next handler after this handler
        :param handler:
        :return: Optional[BlockHandler, None]
        """
        self._next_handler = handler
        return handler

    def handle(self,
               message: Message) -> None:
        """
        Flow chain management method
        check step by step
        :param MessageQueue message: msg to process
        in source have name block from which orchestrator_service get msg
        :return: None
        """
        if not message.get_source():
            # if message don't have source
            # we start from first chain flow
            self.process(message)
        elif message.get_source() == self.name_block:
            if self.post_handler_function:
                # apply post function
                message = self.post_handler_function(message)
            if not message:
                # if not exist message
                return
            if not self._next_handler:
                # last handler in chain
                return
            if self._next_handler.pre_handler_function:
                message = self._next_handler.pre_handler_function(message)
            if not message:
                return
            self._next_handler.process(message)

        elif self._next_handler:
            self._next_handler.handle(message)
        else:
            raise FlowException(f'Not found block for source: {message}')

    def process(self, message: Message):
        """
        Method for executing the logic of a given block
        in it, only send messages to other services
        :param message:
        :return:
        """
        raise NotImplementedError

    def get_list_flow(self) -> str:
        """
        Method return str flow
        :return: str
        """
        if self._next_handler:
            next_blocks = self._next_handler.get_list_flow()
        else:
            next_blocks = 'end'
        return f'{self.name_block} -> {next_blocks}'

    @pre_handler_function.setter
    def pre_handler_function(self, func: Callable = None):
        if not func:
            pass
        elif isinstance(func, (types.FunctionType,
                               types.MethodType)):
            self._pre_handler_function = func
        else:
            raise TypeError('Incorrect type pre_handler_function,'
                            ' the attribute must be a function or None')

    @post_handler_function.setter
    def post_handler_function(self, func: Callable = None):
        if not func:
            pass
        elif isinstance(func, (types.FunctionType,
                               types.MethodType)):
            self._post_handler_function = func
        else:
            raise TypeError('Incorrect type post_handler_function,'
                            ' the attribute must be a function or None')



"""
Module Service with help classes for build service
"""
# pylint: disable=too-few-public-methods
from __future__ import annotations

import logging
import os
import warnings
from typing import Dict, Optional, Union, Tuple

from .exc import *
from .to_extends import CommandHandlerPostStrategy, \
    CommandHandlerStrategy
from ..message import Message

default_logger = logging.Logger(__name__)


class ServiceBlock(object):
    """
    Class ServiceBlock for ServiceBuilder
    """
    __process = None  # type: CommandHandlerStrategy
    __post_process = None  # type: CommandHandlerPostStrategy

    @property
    def process(self):
        """
        process instance CommandHandlerStrategy
        """
        return self.__process

    @property
    def post_process(self):
        """
        post_process instance CommandHandlerPostStrategy
        """
        return self.__post_process

    @process.setter
    def process(self, val: Union[type, CommandHandlerStrategy]):
        """
        process setter
        """
        if isinstance(val, type):
            if getattr(val, '__base__'):
                if val.__base__.__name__ == 'CommandHandlerStrategy':
                    self.__process = val()
                    return
        elif isinstance(val, CommandHandlerStrategy):
            self.__process = val
            return
        raise ServiceBlockException(
            f"`process` object must be of type `CommandHandlerStrategy` and not {type(val)}")

    @post_process.setter
    def post_process(self, val: Union[type, CommandHandlerPostStrategy] = None):
        """
        post_process setter
        """
        if val is None:
            return
        elif isinstance(val, type):
            if getattr(val, '__base__'):
                if val.__base__.__name__ == 'CommandHandlerPostStrategy':
                    self.__post_process = val()
                    return
        elif isinstance(val, CommandHandlerPostStrategy):
            self.__post_process = val
            return

        raise ServiceBlockException(
            f"`post_process` object must be of type "
            f"`CommandHandlerPostStrategy` and not {type(val)}")

    def __init__(self,
                 process: Union[type, CommandHandlerStrategy],
                 post_process: Union[type, CommandHandlerPostStrategy] = None
                 ):
        """
        Init ServiceBlock
        :param CommandHandlerStrategy process:
        :param CommandHandlerPostStrategy post_process:
        """
        self.process = process
        self.post_process = post_process


class ServiceBuilder(object):
    """
    Class for build and aggregate handlers
    """
    _default_post_process = None  # type: CommandHandlerPostStrategy

    @staticmethod
    def _check_default_pp(default_post_process: Union[type, CommandHandlerPostStrategy] = None
                          ) -> Optional[CommandHandlerPostStrategy]:
        if default_post_process is None:
            return None
        elif isinstance(default_post_process, type):
            if getattr(default_post_process, '__base__'):
                if default_post_process.__base__.__name__ == 'CommandHandlerPostStrategy':
                    return default_post_process()

        elif isinstance(default_post_process, CommandHandlerPostStrategy):
            return default_post_process
        msg = f"`default_post_process` object must be of type `CommandHandlerPostStrategy`" \
              f" and not {type(default_post_process)}"
        raise ServiceBlockException(msg)

    def __init__(self, *args,
                 default_post_process: Union[type, CommandHandlerPostStrategy] = None):
        """
        Init ServiceBuilder
        :param List[CommandHandlerStrategy] args:  list CommandHandlerStrategy
         objects for current service
        :param CommandHandlerPostStrategy default_post_process: default post process handler
            if not exist handler for concrete ProcessHandler
        :param kwargs: other args for future
        """

        self._default_post_process = ServiceBuilder._check_default_pp(default_post_process)
        list_blocks = list()
        for block in args:
            try:
                tmp_pp = ServiceBuilder._check_default_pp(block)

                if not self._default_post_process:
                    self._default_post_process = tmp_pp
                    continue
                else:
                    raise DoublePostProcessFunctionDeclaredError
            except ServiceBlockException:
                pass
            if not isinstance(block, ServiceBlock):
                raise TypeError(f"block must be instance class `ServiceBlock`."
                                f" Not {type(block)}")
            list_blocks.append(block)
        if not list_blocks:
            raise EmptyCommandsException
        self._list_blocks = list_blocks

    def build(self,
              log: logging.Logger = None,
              service_instance: Service = None) -> Dict[str:Dict[str:object]]:
        """
        Method build dict handler
        :param logging.Logger log: log application for set into services
        :param Service service_instance: service object
        :return: {'command':{'process': CommandHandlerStrategy,
        'post_process': CommandHandlerPostStrategy}
        """
        if not log:
            log = default_logger
        dict_commands = dict()

        if self._default_post_process:
            self._default_post_process.set_logger(log)
            self._default_post_process.set_service_instance(instance=service_instance)
        for block in self._list_blocks:
            process = block.process
            post_process = block.post_process
            process.set_logger(log)
            process.set_service_instance(instance=service_instance)
            if post_process:
                post_process.set_logger(log)
                post_process.set_service_instance(instance=service_instance)

            else:
                post_process = self._default_post_process

            if not dict_commands.get(process.target_command):
                dict_commands[process.target_command] = {
                    'process': process,
                    'post_process': post_process
                }
            else:
                raise NotUniqueCommandError(f"Command `{process.target_command}`"
                                            f" is not unique for current service")
        return dict_commands


class Service(object):
    """
    Class Service for handle msg-s from queue
    """

    _command_field: str = os.getenv('NameCommandInHandler', 'command')
    _default_command: str = os.getenv('DefaultCommand', 'run')
    _dict_handlers = dict()  # type: Dict[str: CommandHandlerStrategy]
    _is_run_default: bool = False
    _service_commands: ServiceBuilder = None
    is_catch_exceptions: bool = False

    @property
    def service_commands(self) -> Optional[ServiceBuilder]:
        """
        Property _service_commands
        :return: ServiceBuilder
        """
        if not self._service_commands:
            raise NotImplementedError
        return self._service_commands

    @service_commands.setter
    def service_commands(self,
                         service_builder: ServiceBuilder) -> None:
        """
        setter ServiceBuilder
        :param ServiceBuilder service_builder:
        :return: None
        """
        if not isinstance(service_builder, ServiceBuilder):
            raise ServiceBuilderException('Incorrect type `service_builder` variable.')

    def __init__(self,
                 service_builder: ServiceBuilder = None,
                 log: logging.Logger = None,
                 is_run_default: bool = True,
                 command_field: str = 'command',
                 default_command: str = None,
                 is_catch_exceptions: bool = True,
                 ):
        """
        Init Service
        :param ServiceBuilder service_builder: instance builder with handlers command
        :param logging.Logger log: logger
        """
        if is_catch_exceptions:
            self.is_catch_exceptions = True
        if service_builder is None:
            if not isinstance(self.service_commands, ServiceBuilder):
                raise ServiceBuilderException('Incorrect type `service_commands` must be '
                                              'overridden in subclass Service.')
            service_builder = self.service_commands
        elif not isinstance(service_builder, ServiceBuilder):
            raise ServiceBuilderException('Incorrect type `service_builder` variable.')
        if log:
            self.logger = log
        else:
            self.logger = default_logger
        if self._command_field == 'command' and command_field != 'command':
            self._command_field = command_field
        if is_run_default:
            self._is_run_default = True
        if default_command:
            self._default_command = default_command
        # elif self._default_command == 'run' and default_command != 'run':
        #     self._default_command = default_command
        self._dict_handlers = service_builder.build(log=self.logger,
                                                    service_instance=self)
        if self._is_run_default \
                and self._default_command \
                and self._default_command not in self._dict_handlers:
            if len(self._dict_handlers) > 1:
                raise IncorrectDefaultCommand(self._default_command,
                                              list(self._dict_handlers.keys()))

    def _get_handlers(self, msg) -> Tuple[CommandHandlerStrategy, CommandHandlerPostStrategy]:
        command = msg.header.get(self._command_field)
        handler = self._dict_handlers.get(command)  # get handler without keys
        if not handler:
            if len(self._dict_handlers) == 1:
                handler = self._dict_handlers[list(self._dict_handlers.keys())[0]]
            elif self._is_run_default and self._default_command:
                handler = self._dict_handlers.get(self._default_command)
        if not handler:
            raise CommandHandlerNotFoundException(command)
        return handler['process'], handler['post_process']

    def handle(self,
               msg: Message) -> Optional[Message]:
        """
        Method handle msgs from queue for this service
        if only one handler service run this handler
        if don't found handler and exist default command
        :param MessageQueue msg:
        :return:
        """
        try:
            process_handler, post_process_handler = self._get_handlers(msg)
            resp_process = process_handler.process(msg)
            try:
                resp_msg, additional_data = resp_process
            except TypeError:
                resp_msg, additional_data = resp_process, None
            if post_process_handler and resp_msg:
                post_process_handler.post_process(resp_msg,
                                                  additional_data=additional_data)
        except CommandHandlerNotFoundException as exc:
            warnings.warn("deprecated", UnknownCommandWarning)
            self.logger.warning(f"Don't process message. Reason:{exc}", exc_info=True)
            return msg
        except Exception as exc:
            self.logger.warning(str(exc), exc_info=True)
            if not self.is_catch_exceptions:
                raise exc
            return msg

    async def ahandle(self,
                      msg: Message) -> Optional[Message]:
        """
        async Method handle msgs from queue for this service
        if only one handler service run this handler
        if don't found handler and exist default command
        :param MessageQueue msg:
        :return:
        """
        try:
            process_handler, post_process_handler = self._get_handlers(msg)
            resp_process = await process_handler.aprocess(msg)
            try:
                resp_msg, additional_data = resp_process
            except TypeError:
                # if one value
                resp_msg, additional_data = resp_process, None
            except ValueError:
                # if more than one
                resp_msg, *additional_data = resp_process
                if len(additional_data) == 1:
                    additional_data = additional_data[0]
            if post_process_handler and resp_msg:
                await post_process_handler.apost_process(resp_msg,
                                                         additional_data=additional_data)
        except CommandHandlerNotFoundException as exc:
            warnings.warn("deprecated", UnknownCommandWarning)
            self.logger.warning(f"Don't process message. Reason:{exc}", exc_info=True)
            return msg
        except Exception as exc:
            self.logger.warning(str(exc), exc_info=True)
            if not self.is_catch_exceptions:
                raise exc
            return msg



"""
Flow module exceptions
"""


class FlowBlockException(Exception):
    """
    Class custom exception
    for wrong type of flow block
    """

    def __init__(self, message: str = ''):
        self.message = f"Incorrect type block handler. " \
                       f"The block must inherit from the class 'Block'," \
                       f" and not : {message}"
        Exception.__init__(self, self.message)


class FlowBuilderException(Exception):
    """
    Class custom exception
    for wrong types
    """

    def __init__(self, message: str = ''):
        self.message = f'Incorrect type in builder arguments {message}'
        Exception.__init__(self, self.message)



"""
Import exceptions orchestrator_service
"""
# pylint: disable=unused-import,wildcard-import
from .flow.exc import *
from .block.exc import *


class UniqueNameException(Exception):
    """
    Exception for not unique flows
    """
    def __init__(self, not_unique_flow_name: str, _type: str):
        self.message = f'The {_type} name `{not_unique_flow_name}` ' \
                       f'is not unique to this orchestrator_service'
        Exception.__init__(self, self.message)


class NoDateException(Exception):
    """
    Exception if dict flow is empty
    """
    def __init__(self, _type):
        self.message = f'No {_type}s for processing'
        Exception.__init__(self, self.message)


class WrongTypeException(Exception):
    """
    Exception for incorrect inputted types
    """
    def __init__(self, variable: str = 'flows', type_variable: str = 'any'):
        self.message = f'Invalid variable type `{variable}`.' \
                       f' There {variable} should be a list and not `{type_variable}`'



"""
a module with a strategy for what to do with
incorrect messages and a default solution
"""

from abc import ABC, abstractmethod
from typing import Tuple, Any

from orchestrator_service.message import Message


class StrategyIncorrectMessages(ABC):
    """
    class strategy for handling messages of wrong structure
    """

    @property
    def is_process_incorrect_messages(self):
        """
        boolean parameter whether to process invalid messages
        True - process
        flow - ignore
        """
        raise NotImplementedError

    @abstractmethod
    def message_checker(self, message: Message) -> Tuple[bool, Any]:
        """
        Method for validating messages
        :param Message message: message to check
        :return: is_correct_message: bool, warning message
        """
        raise NotImplementedError

    @abstractmethod
    def process_incorrect_message(self, message: Message) -> None:
        """
        method for handling messages with incorrect structure
        save for send report or save to log
        :param Message message: message to process
        :return: None
        """
        raise NotImplementedError


class IgnoreIncorrectMessage(StrategyIncorrectMessages):
    """
    Default class if no strategy for handling invalid messages is specified
    """

    @property
    def is_process_incorrect_messages(self):
        return True

    def message_checker(self, message: Message) -> Tuple[bool, Any]:
        """
        consider all messages correct
        :param Message message: message to check
        :return:
        """
        return True, ''

    def process_incorrect_message(self, message: Message) -> None:
        """
        :return: None
        """
        return



"""
Orchestrator
"""
# pylint: disable=too-few-public-methods, broad-except,too-many-arguments,logging-fstring-interpolation
import inspect
from logging import Logger
from types import ModuleType
from typing import Union, List, Dict, Optional

from .block import Block
from .exc import NoDateException
from .exc import UniqueNameException
from .exc import WrongTypeException
from .flow import Flow
from .utils import StrategyIncorrectMessages, IgnoreIncorrectMessage
from ..message import Message


class Orchestrator:
    """
    Orchestrator class for build service
    """
    _flows = dict()
    _targets = dict()

    def __init__(self,
                 flows: Union[ModuleType, List],
                 blocks: Union[ModuleType, List] = None,
                 incorrect_messages_handler: StrategyIncorrectMessages = None,
                 flow_field: str = 'flow',
                 block_field: str = 'target',
                 log: Logger = None,
                 flows_to_ignore: List[str] = None,
                 blocks_to_ignore: List[str] = None):
        """
        Init Orchestrator
        :param flows:
        :param blocks:
        :param incorrect_messages_handler:
        :param flow_field:
        :param block_field:
        :param log:
        :param List[str] flows_to_ignore: names classes to ignore
        (Attention) applies only if a module flows is passed
        :param List[str] blocks_to_ignore: names classes to ignore
        (Attention) applies only if a module blocks is passed

        """
        if not log:
            log = Logger(__name__)
        self._logger = log

        if incorrect_messages_handler \
                and not isinstance(incorrect_messages_handler,
                                   StrategyIncorrectMessages):
            raise TypeError(f"Incorrect_messages_handler must "
                            f"be a subclass of `StrategyIncorrectMessages`"
                            f"and not {type(StrategyIncorrectMessages)}")
        if not incorrect_messages_handler:
            incorrect_messages_handler = IgnoreIncorrectMessage()
        self._incorrect_messages_handler = incorrect_messages_handler

        if not isinstance(flow_field, str):
            raise TypeError('Flow_field field must be a string')
        if not isinstance(block_field, str):
            raise TypeError('Target_field field must be a string')
        self._flow_field = flow_field
        self._block_field = block_field

        self._flows = self._generate_data(data_to_process=flows,
                                          _type_to_compare=Flow,
                                          attribute_to_get='name_flow',
                                          names_to_ignore=flows_to_ignore,
                                          _type_data='flow')
        if blocks:
            self._targets = self._generate_data(data_to_process=blocks,
                                                _type_to_compare=Block,
                                                attribute_to_get='name_block',
                                                names_to_ignore=blocks_to_ignore,
                                                _type_data='block')

    @staticmethod
    def _generate_data(data_to_process: Union[ModuleType, List],
                       _type_to_compare: type,
                       attribute_to_get: str,
                       names_to_ignore: List[str] = None,
                       _type_data: str = 'flow') -> Dict:
        """

        :param data_to_process:
        :param _type_to_compare:
        :param attribute_to_get:
        :param names_to_ignore:
        :param _type_data:
        :return:
        """
        if not names_to_ignore:
            names_to_ignore = list()

        _data = dict()
        if inspect.ismodule(data_to_process):
            for class_name, clazz in inspect.getmembers(data_to_process,
                                                        inspect.isclass):
                if class_name in names_to_ignore:
                    continue

                if not issubclass(clazz.__base__, _type_to_compare):
                    raise TypeError(f'flow is not inheritor {_type_to_compare.__name__}')
                if _data.get(getattr(clazz, attribute_to_get)):
                    raise UniqueNameException(getattr(clazz, attribute_to_get), _type_data)
                # class is not initialized
                _data[getattr(clazz, attribute_to_get)] = clazz
        elif isinstance(data_to_process, list):
            for flow in data_to_process:
                if isinstance(flow, type):
                    if issubclass(flow.__base__, _type_to_compare):
                        if _data.get(getattr(flow, attribute_to_get)):
                            raise UniqueNameException(getattr(flow, attribute_to_get),
                                                      _type_data)
                        _data[getattr(flow, attribute_to_get)] = flow
                    else:
                        raise TypeError(f'flow: `{flow}` is not'
                                        f' inheritor {_type_to_compare.__name__}')
                elif isinstance(flow, _type_to_compare):
                    if _data.get(getattr(flow, attribute_to_get)):
                        raise UniqueNameException(getattr(flow, attribute_to_get), _type_data)
                    _data[getattr(flow, attribute_to_get)] = flow
                else:
                    raise TypeError(f'flow: `{flow}` is not '
                                    f'inheritor {_type_to_compare.__name__}')
        else:
            raise WrongTypeException('flows', type(data_to_process))
        if not _data:
            raise NoDateException(_type_data)
        return _data

    def handle(self, message: Message) -> Optional[Message]:
        """

        :param Message message: message to process
        :return: message if don't have information for processing or incorrect
        """
        self._logger.debug(str(message))
        is_correct, message = self._incorrect_messages_handler.message_checker(message)
        if not is_correct:
            self._incorrect_messages_handler.process_incorrect_message(message)
            if not self._incorrect_messages_handler.is_process_incorrect_messages:
                return message
        msg_header = message.header
        if self._flow_field in msg_header:
            flow = self._flows.get(msg_header[self._flow_field])  # type: Flow
            if flow:
                if isinstance(flow, type):
                    flow = flow()
                    self._flows[msg_header[self._flow_field]] = flow
                flow.to_go_with_the_flow(message=message)
            else:
                self._logger.error(f'No flow `{msg_header[self._flow_field]}` for processing')
                return message
        elif self._block_field in msg_header:
            target = self._targets.get(msg_header[self._block_field])  # type: Block
            if target:
                if isinstance(target, type):
                    target = target()
                    self._targets[msg_header[self._block_field]] = target
                target.process(message=message)
            else:
                self._logger.error(f'No blocks `{msg_header[self._block_field]}` for processing')
                return message
        else:
            self._logger.error(f"Message don't have information for process {message}")
            return message



"""
Module with classes for inheritance.
And the implementation in these classes of the service logic.
"""
# pylint: disable=too-few-public-methods
import logging
import logging as log_module
from abc import ABC, abstractmethod
from typing import Any, Optional, Tuple

from ..message import Message

DEFAULT_LOGGER = logging.Logger(__name__)


class CommandHandler(ABC):
    """
    Class with method for all handlers
    """
    _logger = DEFAULT_LOGGER  # for one global handler
    _service_instance = None  # single scope for service_commands

    @property
    def logger(self) -> Optional["logger"]:
        """
        get logger
        :return:  logger or None
        """
        return self._logger

    @logger.setter
    def logger(self, val: Any):
        """
        method checks the set value
        :param Any val: value to set
        """
        _val = val
        if not val:
            return
        if type(val) != type:
            val = type(val)
        if val == log_module:
            pass
        elif issubclass(val, log_module.Filterer):
            pass
        else:
            raise TypeError(f"Type must be logger but not {val}")
        self._logger = _val

    def set_logger(self,
                   log: logging.Logger) -> None:
        """
        method set to service global logger if not inited
        :param logging.Logger log: service logger
        :return: None
        """
        self.logger = log

    def set_service_instance(self, instance: "Service") -> None:
        """
        Added a single scope for sharing data in the service
        :param Service instance: service object
        """
        self._service_instance = instance

    def set_to_swap_scope(self,
                          key: str,
                          data: Any) -> bool:
        """
        Method adds a value to the global scope for access from all services
        :param str key: name key
        :param Any data: data to add
        :return: bool: is_added_value
        """
        is_added = False
        if self._service_instance:
            try:
                setattr(self._service_instance, key, data)
                is_added = True
            except Exception:
                if self.logger:
                    self.logger.warning(f'Key `{key}` not added', exc_info=True)
        else:
            if self.logger:
                self.logger.warning("You cann't use swap because it is not initialized ")
        return is_added

    def get_from_swap_scope(self, key: str) -> Optional[Any]:
        """
        Method gets from the global scope a value available for all services
        :param str key: name key to get value
        :return: data if exist or None
        """
        data = None
        if self._service_instance:
            try:
                data = getattr(self._service_instance, key)
            except Exception:
                if self.logger:
                    self.logger.warning('Error while fetching data from swap', exc_info=True)
        else:
            if self.logger:
                self.logger.warning("You cann't use swap because it is not initialized ")
        return data

    def del_from_swap_scope(self,  key: str) -> bool:
        """
        Method removes attribute from swap_scope if was exist
        """
        is_dropped = False
        if self._service_instance:
            try:
                delattr(self._service_instance, key)
                is_dropped = True
            except Exception:
                if self.logger:
                    self.logger.warning(f'Key `{key}` not added', exc_info=True)
        else:
            if self.logger:
                self.logger.warning("You cann't use swap because it is not initialized ")
        return is_dropped



class CommandHandlerStrategy(CommandHandler, ABC):
    """
    handler class
    """

    @property
    def target_command(self):
        """
        this command will determine that the message should be processed by this particular service

        """
        raise NotImplementedError

    @abstractmethod
    def process(self,
                msg: Message) -> Tuple[Message, Optional[Any]]:
        """
        the main method for executing the logic of this handler, must be overridden in the inheritor
        :param MessageQueue msg: msg from queue
        :return: MessageQueue or None if return None post handler will not be called
        """
        raise NotImplementedError

    @abstractmethod
    async def aprocess(self, msg: Message) -> Tuple[Message, Optional[Any]]:
        """
        the main async method for executing the logic of this handler,
        must be overridden in the inheritor
        :param MessageQueue msg: msg from queue
        :return: MessageQueue or None if return None post handler will not be called
        """
        return self.process(msg=msg)


class CommandHandlerPostStrategy(CommandHandler, ABC):
    """
    Post Process Handler
    """

    @abstractmethod
    def post_process(self, msg: Message, additional_data: Optional[Any] = None) -> None:
        """
        method does post processing
        e.g. sending to another queue
        , must be overridden in the inheritor
        :param MessageQueue msg:
        :param additional_data: optional data got from processing block
        :return: None
        """
        raise NotImplementedError

    @abstractmethod
    async def apost_process(self, msg: Message, additional_data: Optional[Any] = None) -> None:
        """
        method does post processing
        e.g. sending to another queue
        , must be overridden in the inheritor
        :param MessageQueue msg:
        :param additional_data: optional data got from processing block

        :return: None
        """
        self.post_process(msg=msg, additional_data=additional_data)



"""
module with default classes
"""
from typing import Optional, Any

from .to_extends import CommandHandlerPostStrategy
from ..message import Message


class DoNothingStrategy(CommandHandlerPostStrategy):
    """
    Default post-processing strategy that does nothing
    """

    def post_process(self, msg: Message, additional_data: Optional[Any] = None) -> None:
        return

    async def apost_process(self, msg: Message, additional_data: Optional[Any] = None) -> None:
        return



"""
Setup module for install lib
"""
import os
from os import path
from typing import List

from setuptools import setup

MODULE_NAME = 'orchestrator_service'
LIB_NAME = 'orchestrator_service'
__version__ = '0.0.8'

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README_PYPI.MD'), encoding='utf-8') as f:
    long_description = f.read()


def get_packages() -> List[str]:
    """
    Help method
    :return: List[str] path to files and folders library
    """
    ignore = ['__pycache__']

    list_sub_folders_with_paths = [x[0].replace(os.sep, '.')
                                   for x in os.walk(MODULE_NAME)
                                   if x[0].split(os.sep)[-1] not in ignore]
    return list_sub_folders_with_paths


setup(name=LIB_NAME,
      version=__version__,
      description='orchestrator_service for microservices architecture',
      author='Denis Shchutkiy',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author_email='denisshchutskyi@gmail.com',
      url='https://github.com/Shchusia/orchestrator',
      packages=get_packages(),
      keywords=['pip', MODULE_NAME],
      )



"""
Service exceptions
"""


# pylint: disable=unnecessary-pass,non-parent-init-called


class ServiceException(Exception):
    """
    Class for Service exceptions
    """
    pass


class ServiceBlockException(BaseException):
    """
    Class for exceptions in ServiceBlock
    """


class ServiceBuilderException(BaseException):
    """
    Class for exceptions in BuilderService
    """
    pass


class NotUniqueCommandError(BaseException):
    """
    Class error if user add the same commands
    """
    pass


class UnknownCommandWarning(Warning):
    """
    Class Warning if handler got msg with incorrect command
    """
    pass


class DoublePostProcessFunctionDeclaredError(ServiceException):
    """
    Exception if many default postprocess handlers
    """

    def __init__(self):
        self.message = "Several postprocessors specified"
        Exception.__init__(self, self.message)


class IncorrectDefaultCommand(ServiceException):
    """
    among the available commands, there is no default command
    """

    def __init__(self, command: str, list_command: list):
        self.message = f"The `{command}` command which is the default " \
                       f"command is not among the valid commands : {str(list_command)}"
        Exception.__init__(self, self.message)


class EmptyCommandsException(ServiceException):
    """
    Empty list of commands for service operation
    """

    def __init__(self, ):
        self.message = "Empty list of commands for service operation"
        Exception.__init__(self, self.message)


class CommandHandlerNotFoundException(ServiceException):
    """
    if not exist handler and not exist default handler
    """

    def __init__(self, command: str):
        self.message = f'No handler for `{command}` command'
        Exception.__init__(self, self.message)



"""
Module with Flow
"""
# pylint: disable=too-few-public-methods
from __future__ import annotations

from typing import Union

from orchestrator_service.message import Message
from .exc import FlowBlockException
from .exc import FlowBuilderException
from ..block import Block


class FlowBlock:
    """
    Block for FlowBuilder
    """
    obj_block: Union[Block, type] = None

    def __init__(self,
                 obj_block,
                 pre_handler_function=None,
                 post_handler_function=None):
        """
        Init FlowBlock
        :param obj_block: type stepBlock
        :param pre_handler_function
        :param  post_handler_function
        """
        try:
            if isinstance(obj_block, type):
                if getattr(obj_block, '__base__'):
                    if obj_block.__base__.__name__ == 'Block':
                        self.obj_block = obj_block

                        self.pre_handler_function = str(pre_handler_function)
                        self.post_handler_function = str(post_handler_function)
                        return

            elif issubclass(type(obj_block), Block):
                self.obj_block = obj_block

                self.pre_handler_function = str(pre_handler_function)
                self.post_handler_function = str(post_handler_function)
                return

            raise FlowBlockException(type(obj_block))
        except FlowBlockException as exc:
            raise exc
        except Exception as exc :
            raise TypeError("Incorrect type `obj_block`")

    def init_block(self, instance_main: Flow) -> Block:
        """
        Method init instance subclass MainBlock
        :param instance_main:
        :return: object subclass MainBlock
        """
        if not isinstance(instance_main, Flow):
            raise TypeError("Value `instance_main` must be a Flow")
        if isinstance(self.obj_block, type):
            self.obj_block = self.obj_block(
                pre_handler_function=getattr(instance_main, self.pre_handler_function, None),
                post_handler_function=getattr(instance_main, self.post_handler_function, None),
            )
        else:
            self.obj_block.pre_handler_function = getattr(instance_main,
                                                          self.pre_handler_function,
                                                          None)
            self.obj_block.post_handler_function = getattr(instance_main,
                                                           self.post_handler_function,
                                                           None)
        return self.obj_block


class FlowBuilder:
    """
    Flow building class
    build chain flow from flow blocks
    """

    def __init__(self,
                 step: FlowBlock,
                 *args):
        """
        Init FlowBuilder
        :param FlowBlock step: first block in flow
        :param List[FlowBlock] args:  other steps  if value exsst
        """
        self.steps = list(args)
        self.steps.insert(0, step)
        for _index, _step in enumerate(self.steps):
            if not isinstance(_step, FlowBlock):
                raise FlowBuilderException(f'on index {_index}')

    def build_flow(self, instance_main: Flow) -> Block:
        """
        Build chain flow for StrategyFlow
        :param instance_main:
        :return:
        """
        flow = self.steps[0].init_block(instance_main)
        cur_step = flow
        for step in self.steps[1:]:
            cur_step = cur_step.set_next(step.init_block(instance_main))
        return flow


class Flow:
    """
    Class for inheritance for a specific flow
    """
    flow_chain: Block = None

    @property
    def name_flow(self):
        """
        Name current flow
        :return: name flow
        """
        raise NotImplementedError

    @property
    def steps_flow(self):
        """
        Steps current flow
        :return:
        """
        raise NotImplementedError

    @steps_flow.setter
    def steps_flow(self,
                   flow: FlowBuilder):
        """
        check the set value to property `steps_flow` value
        :param FlowBuilder flow: builder flow for current flow
        :return: None or exception
        """
        if isinstance(flow, FlowBuilder):
            self.steps_flow = flow
        else:
            raise TypeError('incorrect type flow builder')

    def __init__(self):
        """
        Init Flow
        """

        if isinstance(self.steps_flow, FlowBuilder):
            self.flow_chain = self.steps_flow.build_flow(self)
        else:
            raise TypeError(f"Incorrect type 'steps_flow' - it must be 'FlowBuilder',"
                            f" and not {type(self.steps_flow)}")

    def to_go_with_the_flow(self,
                            message: Message) -> None:
        """
        Method that starts flow execution from the first block
        :param message:
        :return: None
        """
        self.flow_chain.handle(message)

    def get_steps(self) -> str:
        """
        Print steps flow
        :return:
        """
        return self.flow_chain.get_list_flow()



"""
Module with custom msg
"""
from .message import Message


class MessageCustom(Message):
    """
    Class for build custom messages
    """

    def __init__(self, body: dict, header: dict = None):
        """
        Init MessageCustom
        :param dict body:
        :param dict header:
        """
        if header is None:
            header = dict()
        self.body = body  # type: dict
        self.header = header  # type: dict



"""
Importing
"""
from .service import Service
from .service import ServiceBlock
from .service import ServiceBuilder
from .to_extends import CommandHandlerPostStrategy
from .to_extends import CommandHandlerStrategy



"""
Module with Base class for orchestrator_service operation
"""
import json
from pprint import pprint
from typing import Dict, Any, Optional, Union


class Message:
    """
    Class for working with a received message from the queue
    """

    __body = None
    __header = None

    @property
    def body(self):
        """
        Property class message - body
        :return: dict body: message
        """
        return self.__body

    @property
    def header(self):
        """
        Property class message - header
        :return: dict header: message
        """
        return self.__header

    @body.setter
    def body(self, body):
        if not isinstance(body, dict):
            raise TypeError("Body msg must be a dict")
        self.__body = body

    @header.setter
    def header(self, header):
        if not isinstance(header, dict):
            raise TypeError("Header msg must be a dict")
        self.__header = header

    def __init__(self, body: dict, header: dict = None):
        self.body = body
        if header is None:
            header = dict()
        self.header = header

    def __str__(self) -> str:
        return f'<Message body: `{self.__body}` header: `{self.__header}`>'

    def print_message(self) -> None:
        """
        Method print message
        :return: None
        """
        print('Header:')
        pprint(self.__header)
        print("Body")
        pprint(self.__body)

    def update_body(self, date_to_add: Union[Dict, Any],
                    key: Optional[str] = None) -> None:
        """
        Method add value to body
        if a `key` exists, then the date_to_add will be added by this `key`
        if the key doesn't exist then date_to_add must be a dictionary
         and the current __body will be updated
        :param date_to_add:
        :param str key:
        :return: None
        """
        try:
            if key is None:
                if isinstance(date_to_add, dict):
                    self.__body.update(date_to_add)
                else:
                    raise TypeError("date_to_add must be a dictionary if no key is specified")
            else:
                self.__body[key] = date_to_add
        except Exception as exc:
            raise exc

    def update_header(self, date_to_add: Union[Dict, Any],
                      key: Optional[str] = None) -> None:
        """
        Method add value to body
        if a `key` exists, then the date_to_add will be added by this `key`
        if the key doesn't exist then date_to_add must be a dictionary
         and the current __header will be updated
        :param date_to_add:
        :param key:
        :return: None
        """
        try:
            if key is None:
                if isinstance(date_to_add, dict):
                    self.__header.update(date_to_add)
                else:
                    raise TypeError("date_to_add must be a dictionary if no key is specified")
            else:
                self.__header[key] = date_to_add
        except Exception as exc:
            raise exc

    def get_body(self, returned_type_str: bool = True) -> Union[str, Dict]:
        """
        The method returns the body of the message str or dict for future processing
        :param bool returned_type_str: type string may be needed if send to queue
        :return: body in str or dict type
        """
        if returned_type_str:
            return json.dumps(self.__body)
        return self.__body

    def get_header(self, returned_type_str: bool = True) -> Union[str, Dict]:
        """
        The method returns the header of the message str or dict for future processing
        :param bool returned_type_str: type string may be needed if send to rabbit queue
        :return: header in str or dict type
        """
        if returned_type_str:
            return json.dumps(self.__header)
        return self.__header

    def get_source(self) -> str:
        """

        :return:
        """
        return self.__header.get('source', None)

    def set_source(self, new_source: str) -> None:
        """

        :param new_source:
        :return:
        """
        self.__header['source'] = new_source
